import numpy as np
import random
import pickle
from numpy import genfromtxt
from sklearn.linear_model import LinearRegression as LR
from sklearn.linear_model import LassoCV
from sklearn.linear_model import Lasso
from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestRegressor as RF
from sklearn.ensemble import GradientBoostingRegressor as GB
from sklearn import metrics
from sklearn.cluster import SpectralClustering as SC
from sklearn import preprocessing
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from typing import List, Dict, Tuple
from genData import gen0, gen1, gen2, gen3
from method import *
from sklearn.model_selection import KFold
from sklearn.ensemble import VotingRegressor

if __name__ == '__main__':

		# param
		sd = 0.01

		#tuning parameter in the normalization of V matrix
		atune=1

		# num of clusters
		k=2
	  
		# num of modules
		m = 97

		#number of modules in the training set
		m1=64
		
		nrep=1

		# gen data for each module
		c = genfromtxt('slice_localization_data.csv',delimiter=',')
		c=np.delete(c, 0, axis=0)
		n,p=c.shape
		p=p-1
		c[:,1:p]=preprocessing.scale(c[:,1:p])

		#create training and testing index
		la = []
		labs=[]
		Xs = []
		X1s=[]
		X2s=[]
		ys = []
		y1s=[]
		y2s=[]
		ks=[]
		regs = []
		reg1s=[]
		ns=[]  
		ntests=[]
		err1s=[]
		err2s=[]

		#unpooled predictions err3s
		err3s=[]
		for j in range(m):
				random.seed(j+1000)
				la.append(j)
				X,y =c[c[:,0]==j,:-1],c[c[:,0]==j,-1]
				Xs.append(X)
				ys.append(y)
				ns.append(X.shape[0])

				#for the test sets, each module in the test set is divided into two parts, the first part
				# is for clustering, the second part is for validation
				index1=np.sort(random.sample(range(X.shape[0]),X.shape[0]//2))
				index2=np.sort(list(set(range(X.shape[0])) - set(index1)))
				X1s.append(X[index1,:])
				X2s.append(X[index2,:])
				ntests.append(X[index2,:].shape[0])
				y1s.append(y[index1])
				y2s.append(y[index2])
				reg0 = RF(n_estimators=50, max_depth=4)
				#reg = GB(n_estimators=100, max_depth=5)
				reg1 = Lasso(alpha=1)
				scores = cross_val_score(reg0, X, y, cv=5)
				scores1 = cross_val_score(reg1, X, y, cv=5)
				if np.mean(scores) > np.mean(scores1):
						reg= reg0
				else:
						reg= reg1
				reg.fit(X, y)
				scores_2 = cross_val_score(reg0, X[index1,:], y[index1], cv=5)
				scores1_2 = cross_val_score(reg1, X[index1,:], y[index1], cv=5)
				if np.mean(scores_2) > np.mean(scores1_2):
						rego= reg0
				else:
						rego= reg1
				rego.fit(X[index1,:], y[index1])
				regs.append(reg)
				reg1s.append(rego)


		print('\ntrue labels are: \n', la)

		# validation error matrix
		V = np.zeros((m, m))
		for i in range(m):
				for j in range(m):
						y_pred = regs[i].predict(Xs[j])
						V[i, j] = metrics.mean_squared_error(ys[j], y_pred)    
		V=atune*(V)/np.std(V)

		# weight matrix
		# not yet normalize the weight to sum to one
		W = np.zeros((m, m))
		for i in range(m):
				for j in range(m):
						W[i, j] = np.abs(V[i, j] - V[j, j]) + np.abs(V[j, i] - V[i, i])
		W = np.exp(-W)  # transform to 'SIMILARITY'

		for i in range(nrep):               
				random.seed(i+100)
				IndexTrain=np.sort(random.sample(range(m),m1))
				IndexTest=np.sort(list(set(range(m)) - set(IndexTrain)))
				train=np.in1d(c[:,0],IndexTrain)
				test=np.logical_not(train)           
				Xtrain,ytrain =c[train,:-1], c[train,-1]
				Xtest=np.vstack([X2s[i] for i in IndexTest])
				ytest=np.hstack([y2s[i] for i in IndexTest])
				#Xtest,ytest =c[test,:-1], c[test,-1]
				# getClust(W, 3)  # this function is not in use
				lab, k = getClust_custom(W[IndexTrain][:,IndexTrain], k=None, PRINT=False)
				labs.append(lab)
				ks.append(k)
				Wnew=np.zeros((k,m-m1))

				#decide the which cluster each module in the test set belongs to
				V2 = np.zeros((m, m))
				for s in range(m):
						for j in range(m):
								if (s in IndexTrain) & (j in IndexTest): 
									y_pred = regs[s].predict(X1s[j])
									V2[s, j] = metrics.mean_squared_error(y1s[j], y_pred)
								if (s in IndexTrain) & (j in IndexTrain):
									y_pred = regs[s].predict(Xs[j])
									V2[s, j] = metrics.mean_squared_error(ys[j], y_pred)
								if (s in IndexTest) & (j in IndexTest):
									y_pred = reg1s[s].predict(X1s[j])
									V2[s, j] = metrics.mean_squared_error(y1s[j], y_pred)
								if (s in IndexTest) & (j in IndexTrain):
									y_pred = reg1s[s].predict(Xs[j])
									V2[s, j] = metrics.mean_squared_error(ys[j], y_pred)
								# Mean absolute error regression loss
								# validation of i on data j
								# V[i, j] = metrics.mean_absolute_error(ys[j], y_pred)
								# Mean squared error regression loss
									
				V2=atune*(V2)/np.std(V2)
		
				# weight matrix
				# not yet normalize the weight to sum to one
				W2 = np.zeros((m, m))
				for s in range(m):
						for j in range(m):
								#aij= ns[j]*(V[j, i] + V[j, j])
								#bij= ns[i]*(V[j, i] + V[j, j])
								#aaij=aij/(aij+bij)
								#W[i, j] = aaij*np.abs(V[i, j] - V[j, j]) + (1-aaij)*np.abs(V[j, i] - V[i, i])
								W2[s, j] = np.abs(V2[s, j] - V2[j, j]) + np.abs(V2[j, s] - V2[s, s])
				W2 = np.exp(-W2)                
				for j in range(k):
						Wnew[j,:]=np.mean(W2[IndexTrain[lab==j]][:,IndexTest], axis=0)                     
				#get the cluster of the test module (the largest similarity)
				labtest=np.argmax(Wnew,axis=0)
				#not cluster
				reg0 = RF(n_estimators=50, max_depth=4)
				reg0.fit(Xtrain, ytrain)
				y_pred = reg0.predict(Xtest)
				err1=metrics.mean_squared_error(ytest, y_pred)
				err1s.append(err1)
				regpool=[]
				regunpool1=[]
				for t in range(k):
						clusterindex=np.in1d(c[:,0],IndexTrain[lab==t])
						reg1=RF(n_estimators=50, max_depth=4)
						reg1.fit(c[clusterindex,:-1], c[clusterindex,-1])
						regpool.append(reg1)
						regunpool=[]
						for tt in range(len(IndexTrain[lab==t])):
								clusterindexx=np.in1d(c[:,0],IndexTrain[lab==t][tt])
								regg=RF(n_estimators=50, max_depth=4)
								regg.fit(c[clusterindexx,:-1], c[clusterindexx,-1])
								regunpool.append(regg)
						regunpool1.append(regunpool)
				errs=[]
				errs1=[]
				nss=[]                
				nstest=np.array(ntests)[np.array(IndexTest)]
				for t in range(len(np.unique(labtest))):
						labtest1=np.repeat(labtest,nstest,axis=0)
						Xtestcluster=Xtest[labtest1==np.unique(labtest)[t],:]
						y_pred=regpool[np.unique(labtest)[t]].predict(Xtestcluster)
						y_pred1=np.mean([ii.predict(Xtestcluster) for ii in regunpool1[np.unique(labtest)[t]]],axis=0)
						n=Xtestcluster.shape[0]
						ytestcluster=ytest[labtest1==np.unique(labtest)[t]]
						errs0=metrics.mean_squared_error(ytestcluster, y_pred)
						errs01=metrics.mean_squared_error(ytestcluster, y_pred1)
						errs.append(errs0)
						errs1.append(errs01)
						nss.append(n)
				err2=np.sum(np.multiply(errs,nss))/np.sum(nss)
				err2s.append(err2)
				err3=np.sum(np.multiply(errs1,nss))/np.sum(nss)
				err3s.append(err3)

		errdiff=np.array(err1s)-np.array(err2s)
		my_obj=object()
		my_var=(la,ks,ns,err1s,err2s,errdiff,err3s)

		filename = "experiments__slice_new_rep100.pickle"
		with open(filename, 'wb') as f: #save data
				pickle.dump((my_obj, my_var), f)
