import numpy as np
import random
import pickle
from numpy import genfromtxt
from sklearn.linear_model import LinearRegression as LR
from sklearn.linear_model import LassoCV
from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestRegressor as RF
from sklearn.ensemble import GradientBoostingRegressor as GB
from sklearn import metrics
from sklearn.cluster import SpectralClustering as SC
from sklearn import preprocessing
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from typing import List, Dict, Tuple
from method import *


if __name__ == '__main__':

        # param
        sd = 0.01

        #tuning parameter in the normalization of V matrix
        atune=1

        # num of clusters
        k=2

        # num of modules
        m = 50
        
        nrep=100
        nsplit=5

        # gen data for each module
        c = genfromtxt('Data_for_UCI_named.csv',delimiter=',')
        c=np.delete(c, 0, axis=0)
        c=np.delete(c, -1, axis=1)
        n,p=c.shape
        p=p-1
        c[:,0:(p-1)]=preprocessing.scale(c[:,0:(p-1)])

        #create training and testing index
        my_vars=[]
        for jj in range(nsplit):
            la = []
            las=[]
            ks=[]
            ns=[]

            # store the error where there is attack   and we do clustering   
            err2s=[]

            # store the error where there is attack and we know exactly which are attacked (oracle)
            erros=[]

            #store thr error when colloaborating with all others
            erras=[]

            #divide the data into ntrain=8000 and ntest=2000
            ntrain=8000
            ntetst=2000
            random.seed(1234+jj)
            IndexTrain=np.sort(random.sample(range(n),ntrain))
            IndexTest=np.sort(list(set(range(n)) - set(IndexTrain)))      
            ctrain=c[IndexTrain,:]
            ctest=c[IndexTest,:]
            Xtrain=ctrain[:,:-1]
            ytrain=ctrain[:,-1]
            Xtest=ctest[:,:-1]
            ytest=ctest[:,-1]

            #within the training set, create the label for each observation
            #we divide the 8000=160X50, 50 is # of modules  
            indexall=np.repeat(range(m),np.repeat(160,50),axis=0)
            
            #err0 for prediction MSE under no attack at all
            #reg0 = RF(n_estimators=50, max_depth=4)
            reg0=LR()
            reg0.fit(Xtrain, ytrain)
            y_pred = reg0.predict(Xtest)
            err0=metrics.mean_squared_error(ytest, y_pred)
                   
            #err1 for prediction MSE using only one module that is known unattacked
            Xtrain1=ctrain[indexall==(m-1),:-1]
            ytrain1=ctrain[indexall==(m-1),-1]
            reg0=LR()
            reg0.fit(Xtrain1, ytrain1)
            y_pred = reg0.predict(Xtest)
            err1=metrics.mean_squared_error(ytest, y_pred)

            # severity of attack
            for sev in range(m):

            #if no attack, use the whole training set to build a model
            #calculate the simularity matrix for the 50 modules in the training set
                ys = []
                Xs=[]
                regs = []
                for j in range(m):
                        la.append(j)
                        X,y0 =np.copy(ctrain[indexall==j,:-1]),np.copy(ctrain[indexall==j,-1])
                        #y0=np.copy(y)
                        if j < sev :
                            y=y0*(-1)
                        else:
                                y=y0
                        Xs.append(X)
                        ys.append(y)
                        ns.append(X.shape[0])
                        reg0 = RF(n_estimators=50, max_depth=4)
                        #reg = GB(n_estimators=100, max_depth=5)
                        reg1 = LR()
                        random.seed(111+j)
                        scores = cross_val_score(reg0, X, y, cv=5)
                        random.seed(111+j)
                        scores1 = cross_val_score(reg1, X, y, cv=5)
                        if np.mean(scores) > np.mean(scores1):
                                reg= reg0
                        else:
                                reg= reg1
                        reg.fit(X, y)
                        regs.append(reg)
        
                print('\ntrue labels are: \n', la)
        
                # validation error matrix
                V = np.zeros((m, m))
                for i in range(m):
                        for j in range(m):
                                y_pred = regs[i].predict(Xs[j])
                                # Mean absolute error regression loss
                                # validation of i on data j
                                # V[i, j] = metrics.mean_absolute_error(ys[j], y_pred)
                                # Mean squared error regression loss
                                V[i, j] = metrics.mean_squared_error(ys[j], y_pred)    
                V=atune*(V)/np.std(V)
        
                # weight matrix
                # not yet normalize the weight to sum to one
                W = np.zeros((m, m))
                for i in range(m):
                        for j in range(m):
                                aij= ns[j]*(V[j, i] + V[i, i])
                                bij= ns[i]*(V[i, j] + V[j, j])
                                aaij=aij/(aij+bij)
                                W[i, j] = aaij*np.abs(V[i, j] - V[j, j]) + (1-aaij)*np.abs(V[j, i] - V[i, i])
                W = np.exp(-W)  # transform to 'SIMILARITY'
                lab,k = getClust_custom(W, k=None, PRINT=False)
                print(lab)
                las.append(lab)
                ks.append(k)

                #using the unattacked part of the training set to build model
                lab1=np.repeat(lab,np.repeat(160,50),axis=0)
                Xtrainpool=ctrain[lab1==lab[m-1],:-1]
                ytrainpool0=[ys[index] for index in np.arange(m)[lab==lab[m-1]]]
                ytrainpool=np.asarray(ytrainpool0).ravel()
                reg0=LR()
                reg0.fit(Xtrainpool, ytrainpool)
                y_pred = reg0.predict(Xtest)
                err2=metrics.mean_squared_error(ytest, y_pred)
                err2s.append(err2)

                #oracle
                Xoracle=ctrain[indexall>=sev,:-1]
                yoracle=ctrain[indexall>=sev,-1]
                reg1=LR()
                reg1.fit(Xoracle, yoracle)
                y_pred = reg1.predict(Xtest)
                erro=metrics.mean_squared_error(ytest, y_pred)
                erros.append(erro)

                #collaborate with all
                reg3=LR()
                reg3.fit(Xtrain,np.asarray(ys).ravel())
                y_pred = reg3.predict(Xtest)
                erra=metrics.mean_squared_error(ytest,y_pred)
                erras.append(erra)
            
            my_obj=object()
            my_var=(las,ks,ns,err0,err1,err2s,erros,erras)
            my_vars.append(my_var)

        filename = "/experiments__egs1_newest.pickle"
        with open(filename, 'wb') as f: #save data
                pickle.dump(my_vars, f)
