import numpy as np
import pickle
from sklearn.linear_model import LinearRegression as LR
from sklearn.linear_model import LassoCV
from sklearn.linear_model import Lars 
from sklearn import model_selection
from sklearn.ensemble import RandomForestRegressor as RF
from sklearn.ensemble import GradientBoostingRegressor as GB
from sklearn import metrics
from sklearn.cluster import SpectralClustering as SC
from sklearn import preprocessing
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from typing import List, Dict, Tuple
from genData import gen0, gen1, gen2, gen3
from method import *
from sklearn.neighbors import KNeighborsRegressor as KNN
from sklearn.svm import SVR
from sklearn.neural_network import MLPRegressor as NN
from sklearn.linear_model import Ridge
from sklearn.linear_model import Lasso
from sklearn.linear_model import ElasticNet as EN

if __name__ == '__main__':

	# param
	n = 100
	p = 1000
	sd = 0.01
	nrep=100

	#number of candidate methods
	ncandi=9

	# num of clusters
	k = 2

	# multiplier
	c = 10

	# num of modules
	m = c * k

	# label of cluster
	la = []
	labs=[]
	ks=[]
	Xs = []
	ys = []
	regs = []
	names=[]
	namelist=['RF','KNN','SVR','NN','GB','Lasso','Lars','EN','Ridge']
	ns=(np.arange(20)+1)*100
	IDs=[]
	for s in range(nrep):
		for t in range(5):
		# gen data for each module
			for i in range(k):
				for j in range(c):
					la.append(i)
					X, y = eval('gen'+str(i+2)+'(n, p, sd)')
					X = preprocessing.scale(X)
					y = preprocessing.scale(y)
					Xs.append(X)
					ys.append(y)
					reg0 = RF(n_estimators=50, max_depth=4)
					#reg = GB(n_estimators=100, max_depth=5)
					reg1 = KNN(n_neighbors=2)
					reg2 = SVR(gamma='scale', C=1.0, epsilon=0.2)
					reg3 = NN()
					reg4 = GB(n_estimators=50, max_depth=4)
					reg5 = Lasso(alpha=0.1)
					reg6 = Lars(n_nonzero_coefs=1)
					reg7 = EN()
					reg8 = Ridge(alpha=1.0)
					index=np.array([i for i in range(t)]+[range(4,9)[i] for i in range(t+1)])
					
					results = []
					for model in index:
						kfold = model_selection.KFold(n_splits=10,shuffle=True, random_state=(7+t))
						cv_results = model_selection.cross_val_score(eval('reg'+str(model)), X, y, cv=kfold, scoring='neg_mean_squared_error')
						results.append(cv_results.mean())
					best1=np.argmax(np.array(results))
					best=index[best1]
					reg=eval('reg'+str(best))
					name=namelist[best]
					reg.fit(X,y)
					regs.append(reg)
					names.append(name)
					IDs.append(best)
			print('\ntrue labels are: \n', la)

		# validation error matrix
			V = np.zeros((m, m))
			for i in range(m):
				for j in range(m):
					y_pred = regs[i].predict(Xs[j])
				# Mean absolute error regression loss
				# validation of i on data j
				# V[i, j] = metrics.mean_absolute_error(ys[j], y_pred)
				# Mean squared error regression loss
					V[i, j] = metrics.mean_squared_error(ys[j], y_pred)	
			V=V/np.std(V)        
		# weight matrix
		# not yet normalize the weight to sum to one
			W = np.zeros((m, m))
			for i in range(m):
				for j in range(m):
					W[i, j] = np.abs(V[i, j] - V[j, j]) + np.abs(V[j, i] - V[i, i])
			W = np.exp(-W)  # transform to 'SIMILARITY'

		# getClust(W, 3)  # this function is not in use
			lab, k = getClust_custom(W, k=None, PRINT=False)
			labs.append(lab)
			ks.append(k)
	my_obj=object()
	my_var=(labs,la,ks,ns,regs,names,IDs)
	filename = "experiments_robust_rep100.pickle"

	with open(filename, 'wb') as f: #save data
			pickle.dump((my_obj, my_var), f)
