import numpy as np 
from sklearn import datasets
import matplotlib.pyplot as plt
import sys
import random


def gen0(n, p, s, PRINT=False):
	# feature matrix
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1, 2, 3, 4]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y

def gen1(n, p, s, PRINT=False):
	# feature matrix
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [3, -2, 1, 2]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y




def gen2(n, p, s, PRINT=False):
	if p < 4:
		sys.exit('p not large enough in gen1()!')
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	X[:, :4], y = datasets.make_friedman2(n_samples=n, noise=s)
	if PRINT:
		f = plt.figure(2)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y


def gen3(n, p, s, PRINT=False):
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	X[:, :4], y = datasets.make_friedman3(n_samples=n, noise=s)
	if PRINT:
		f = plt.figure(3)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y 

#simulation 2: simulation 1 with high noise sigma=s
def gen4(n, p, s, PRINT=False):
	# feature matrix
	X = np.random.normal(loc=0.0, scale=1.0, size=(n,p))
	# noise
	v = np.random.normal(loc=0.0, scale=s, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1, 2, 3, 4]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y

def gen5(n, p, s, PRINT=False):
	# feature matrix
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	# noise
	v = np.random.normal(loc=0.0, scale=s, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [3, -2, 1, 2]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y


#simulation 3: simulation 1 with high correlation
def gen6(n, p, s, PRINT=False):
	# feature matrix 
	mean = np.zeros(p)
	phi = .5
	t = np.arange(p)+1
	cov = phi**abs(np.subtract.outer(t,t))
	X = np.random.multivariate_normal(mean, cov, n)
	# noise
	v = np.random.normal(loc=0.0, scale=s, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1, 2, 3, 4]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y

def gen7(n, p, s, PRINT=False):
	# feature matrix
	mean = np.zeros(p)
	phi = .5
	t = np.arange(p)+1
	cov = phi**abs(np.subtract.outer(t,t))
	X = np.random.multivariate_normal(mean, cov, n)
	# noise
	v = np.random.normal(loc=0.0, scale=s, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [3, -2, 1, 2]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y
#simulation 4: one cluster with high correlation
def gen8(n, p, s, PRINT=False):
	# feature matrix 
	mean = np.zeros(p)
	phi = .9
	t = np.arange(p)+1
	cov = phi**abs(np.subtract.outer(t,t))
	X = np.random.multivariate_normal(mean, cov, n)
	# noise
	v = np.random.normal(loc=0.0, scale=s, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1, 2, 3, 4]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y

def gen9(n, p, s, PRINT=False):
	# feature matrix
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	# noise
	v = np.random.normal(loc=0.0, scale=s, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [3, -2, 1, 2]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y

#simulation 5: one cluster with high correlation
def gen10(n, p, s, PRINT=False):
	# feature matrix 
	mean = np.zeros(p)
	phi = .9
	t = np.arange(p)+1
	cov = phi**abs(np.subtract.outer(t,t))
	X = np.random.multivariate_normal(mean, cov, n)
	# noise
	v = np.random.normal(loc=0.0, scale=5.0, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1, 2, 3, 4]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y

def gen11(n, p, s, PRINT=False):
	# feature matrix
	mean = np.zeros(p)
	phi = .9
	t = np.arange(p)+1
	cov = phi**abs(np.subtract.outer(t,t))
	X = np.random.multivariate_normal(mean, cov, n)
	# noise
	v = np.random.normal(loc=0.0, scale=5.0, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [3, -2, 1, 2]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y

def gen12(M,n,s, p=1,PRINT=False):
	# feature matrix
	ID=np.repeat(range(M),np.repeat(n,M))
	X = np.random.normal(loc=0.0, scale=1.0, size=(n*M, p))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n*M)
	r= np.sort(np.random.uniform(low=1, high=10, size=n*M))
	# coefficient
	beta = np.zeros(p)
	beta[:1] = [1]
	# observation
	feta=X.dot(beta)
	y=np.sign(feta) * (np.abs(feta))** (r/2)+v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y, ID,r	

def gen13(n, p, s, PRINT=False):
	# feature matrix
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1, 2, 3, 4]
	# observation
	y = X[:,0] + X[:,1]+X[:,2]+X[:,3]+ v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y
def gen14(n, p, s, PRINT=False):
	# feature matrix
	X = np.random.normal(loc=0.0, scale=1.0, size=(n, p))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n)
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1, 2, 3, 4]
	# observation
	y = X.dot(beta) + v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y
#one continuous sensitive attribute that is independent of X
def gen15(M,n,s, p,PRINT=False):
	# feature matrix
	ID=np.repeat(range(M),np.repeat(n,M))
	X = np.random.normal(loc=0.0, scale=1.0, size=(n*M, p))
	S1= np.random.normal(loc=0.0, scale=1.0, size=(M, 1))
	S=np.repeat(S1,np.repeat(n,M))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n*M)
	r= np.sort(np.random.uniform(low=1, high=10, size=n*M))
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1,2,-2,2]
	# observation
	feta=X.dot(beta)+s*S
	y=feta+v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y, ID,r, S
#one continuous sensitive attribute that is independent of X
def gen156(M,n,s, p,PRINT=False):
	# feature matrix
	ID=np.repeat(range(M),np.repeat(n,M))
	X = np.random.normal(loc=0.0, scale=1.0, size=(n*M, p))
	S= np.random.normal(loc=0.0, scale=1.0, size=(n*M, 1)).flatten()
	#S=np.repeat(S1,np.repeat(n,M))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n*M)
	r= np.sort(np.random.uniform(low=1, high=10, size=n*M))
	# coefficient
	beta = np.zeros(p)
	beta[:4] = [1,2,-2,2]
	# observation
	feta=X.dot(beta)+s*S
	y=feta+v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y, ID,r, S
#one categorical sensitive attribute that is independent of X
def gen155(M,n,s,group,p,PRINT=False):
    # feature matrix
    ID=np.repeat(range(M),np.repeat(n,M))
    X = np.random.normal(loc=0.0, scale=1.0, size=(n*M, p))
    S1= [(k+group-1)%group for k in range(1,M+1)]
    S=np.repeat(S1,np.repeat(n,M))
    # noise
    v = np.random.normal(loc=0.0, scale=1.0, size=n*M)
    r= np.sort(np.random.uniform(low=1, high=10, size=n*M))
    # coefficient
    beta = np.zeros(p)
    beta[:4] = [1,2,-2,2]
    # observation
    feta=X.dot(beta)+s*S
    y=feta+v
    if PRINT:
        f = plt.figure(0)
        plt.scatter(X[:, 0], y)
        plt.title('y')
        plt.xlabel('')
        plt.show()

    return X, y, ID,r, S
def gen16(M,n,s, p=1,PRINT=False):
	# feature matrix
	ID=np.repeat(range(M),np.repeat(n,M))
	X = np.random.normal(loc=0.0, scale=1.0, size=(n*M, p))
	# noise
	v = np.random.normal(loc=0.0, scale=1.0, size=n*M)
	r= np.sort(np.random.uniform(low=1, high=5, size=n*M))
	# coefficient
	beta = np.zeros(p)
	beta[:1] = [1]
	# observation
	feta=X.dot(beta)
	y=np.sign(feta) * (np.abs(feta))** (r/2)+v
	if PRINT:
		f = plt.figure(0)
		plt.scatter(X[:, 0], y)
		plt.title('y')
		plt.xlabel('')
		plt.show()

	return X, y, ID,r	

if __name__ == '__main__':
	# param
	n = 100
	p = 10
	s = 1
	# X, y = gen0(n, p, s, True)
	X, y = gen1(n, p, s, True)
	# X, y = gen2(n, p, s, True)
	X, y = gen3(n, p, s, True)
